﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.Linq;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using ThinkGeo.MapSuite.WpfDesktopEdition;
using ThinkGeo.MapSuite.Core;
using WmkRenderLayer;

namespace EMS.ThinkGeoLibrary.MapSuite.Controls
{
    public class ExtendedWpfMap : WpfMap
    {
        /// <summary>
        /// Occurs when drawing progress changed.
        /// </summary>
        public event EventHandler<DrawingProgressChangedExtendedWpfMapEventArgs> DrawingProgressChanged;

        public ExtendedWpfMap()
            : base()
        {
            Overlays.CollectionChanged += Overlays_CollectionChanged;
            this.ComputeCurrentProgressUsingBuffer = true;
        }

        /// <summary>
        /// Set this to true if you want draw progress to
        /// include progress of buffer tiles.  Default is true.
        /// Setting this property to false requires an expensive 
        /// Intersects operation to find out which tiles are within 
        /// visible map extent.  While this works, it is not recommended.
        /// </summary>
        public bool ComputeCurrentProgressUsingBuffer { get; set; }

        /// <summary>
        /// Set this to true to prevent map from zooming in on double click.
        /// </summary>
        public bool PreventZoomOnDoubleClick { get; set; }


        protected override void OnMapDoubleClick(MapClickWpfMapEventArgs e)
        {
            if (!PreventZoomOnDoubleClick)
                base.OnMapDoubleClick(e);
        }

        protected virtual void OnDrawingProgressChanged(DrawingProgressChangedExtendedWpfMapEventArgs e)
        {
            if (DrawingProgressChanged != null)
            {
                DrawingProgressChanged(this, e);
            }
        }

        private void Overlays_CollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
        {
            switch (e.Action)
            {
                case NotifyCollectionChangedAction.Add:
                    int newCount = e.NewItems.Count;
                    TileOverlay tileOverlay = null;
                    foreach (Overlay overlay in e.NewItems)
                    {
                        tileOverlay = overlay as TileOverlay;
                        if (tileOverlay != null && tileOverlay.TileType != TileType.SingleTile)
                            tileOverlay.DrawTilesProgressChanged += tileOverlay_DrawTilesProgressChanged;
                        else
                        {
                            if (newCount == 1 && tileOverlay != null)
                            {
                                tileOverlay.Drawn += OnSingleTileDrawn;
                            }
                        }
                    }
                    break;
                case NotifyCollectionChangedAction.Remove:
                    newCount = e.OldItems.Count;
                    foreach (Overlay overlay in e.OldItems)
                    {
                        tileOverlay = overlay as TileOverlay;
                        if (tileOverlay != null && tileOverlay.TileType != TileType.SingleTile)
                            tileOverlay.DrawTilesProgressChanged -= tileOverlay_DrawTilesProgressChanged;

                        tileOverlay.Drawn -= OnSingleTileDrawn;
                        var layerO = overlay as LayerOverlay;
                        if (layerO != null)
                        {
                            foreach (var layer in layerO.Layers)
                            {
                                layer.IsVisible = false;
                                layer.Close();
                            }

                            layerO.Layers.Clear();
                        }

                        overlay.Dispose();
                    }
                    break;
                default:
                    break;
            }
        }

        void OnSingleTileDrawn(object sender, DrawnOverlayEventArgs e)
        {
            OnDrawingProgressChanged(new DrawingProgressChangedExtendedWpfMapEventArgs(100, null));
        }

        int previousPercentage = 0;
        private void tileOverlay_DrawTilesProgressChanged(object sender, DrawTilesProgressChangedTileOverlayEventArgs e)
        {
            Dispatcher.BeginInvoke(new Action(() => { OnTileOverlayProgressChanged(sender, e); }));
        }

        private void OnTileOverlayProgressChanged(object sender, DrawTilesProgressChangedTileOverlayEventArgs e)
        {
            TileOverlay currentTileOverlay = (TileOverlay)sender;
            currentTileOverlay.OverlayCanvas.Tag = e.ProgressPercentage;

            IEnumerable<TileOverlay> tileOverlays = Overlays.OfType<TileOverlay>()
                .Where((to => to.TileType != TileType.SingleTile))
                .Where(to => to.OverlayCanvas.Tag != null);

            if (ComputeCurrentProgressUsingBuffer)
                ComputeCurrentDrawProgressUsingBuffer(tileOverlays);
            else
                ComputeCurrenDrawProgressWithoutBuffer(tileOverlays);
        }

        private void ComputeCurrenDrawProgressWithoutBuffer(IEnumerable<TileOverlay> tileOverlays)
        {
            var tilesInVisibleExtent = tileOverlays.SelectMany(o => ((Canvas)o.OverlayCanvas.Children[0]).Children
                                                           .OfType<ThinkGeo.MapSuite.WpfDesktopEdition.Tile>()
                                                           .Where(tile => o.MapArguments.CurrentExtent.Intersects(tile.TargetExtent)));

            int total = tilesInVisibleExtent.Count();
            int currentProgress = 100;

            if (total != 0)
            {
                var tilesLoaded = tilesInVisibleExtent.Where(tile => tile.IsOpened);
                currentProgress = tilesLoaded.Count() * 100 / total;
            }

            if (Math.Abs(currentProgress - previousPercentage) >= 5 || currentProgress == 100 && previousPercentage != 100)
            {
                previousPercentage = currentProgress;
                OnDrawingProgressChanged(new DrawingProgressChangedExtendedWpfMapEventArgs(currentProgress, null));
            }
        }

        private void ComputeCurrentDrawProgressUsingBuffer(IEnumerable<TileOverlay> tileOverlays)
        {
            int total = tileOverlays.Count();
            int currentProgress = 100;
            if (total != 0)
            {
                int progress = 0;
                foreach (TileOverlay tileOverlay in tileOverlays)
                {
                    if (tileOverlay.OverlayCanvas.Tag != null)
                    {
                        var overlayProgress = (int)tileOverlay.OverlayCanvas.Tag;
                        progress += overlayProgress;
                    }
                }

                currentProgress = progress / total;

                if (Math.Abs(currentProgress - previousPercentage) >= 5 || currentProgress == 100 && previousPercentage != 100)
                {
                    previousPercentage = currentProgress;
                    OnDrawingProgressChanged(new DrawingProgressChangedExtendedWpfMapEventArgs(currentProgress, null));
                }
            }
        }

        #region Map extensions
        /// <summary>
        /// Call this method to partition the default zoom levels.  Method will 
        /// return a ZoomLevelSet based on the default ZoomLevel and the parameters you pass.
        /// </summary>
        /// <param name="zoomLevelToStartStepping">Specifiy 0 for first zoom level and 20 for last zoom level.</param>
        /// <param name="stepsBetweenZoomLevels">Number of steps in between zoom levels.  For example, specifying 2
        /// for this paramter means we will derive 2 zoom levels between the default zoom level.</param>
        /// <returns></returns>
        public ZoomLevelSet PartitionZoomLevelSet(int zoomLevelToStartStepping, int stepsBetweenZoomLevels)
        {
            return ZoomLevelSet = PartitionZoomLevelSet(zoomLevelToStartStepping, stepsBetweenZoomLevels, ZoomLevelSet);
        }

        public ZoomLevelSet PartitionZoomLevelSet(int zoomLevelToStartStepping,
            int stepsBetweenZoomLevels, ZoomLevelSet referenceZoomLevelSet)
        {
            ZoomLevelSet partitionedZoomLevelSet = new ZoomLevelSet();
            Collection<ZoomLevel> zoomLevels = referenceZoomLevelSet.GetZoomLevels();
            partitionedZoomLevelSet.CustomZoomLevels.Add(new ZoomLevel(zoomLevels[0].Scale));
            ZoomLevel zoomLevel = null;
            for (int index = 0; index < zoomLevelToStartStepping; index++)
            {
                zoomLevel = zoomLevels[index];
                partitionedZoomLevelSet.CustomZoomLevels.Add(zoomLevel);
            }

            for (int index = zoomLevelToStartStepping; index < zoomLevels.Count - 1; index++)
            {
                zoomLevel = zoomLevels[index];
                double lowerScale = ZoomLevelSet.GetLowerZoomLevelScale(zoomLevel.Scale, referenceZoomLevelSet);
                if (lowerScale != zoomLevel.Scale)
                {
                    for (int x = 1; x < stepsBetweenZoomLevels + 2; x++)
                    {
                        double steppedScale = zoomLevel.Scale - (((zoomLevel.Scale - lowerScale) / (stepsBetweenZoomLevels + 1)) * x);
                        partitionedZoomLevelSet.CustomZoomLevels.Add(new ZoomLevel(steppedScale));
                    }
                }
            }
            return partitionedZoomLevelSet;
        }

        public bool ContainsLayer(string layerName)
        {
            return this.Overlays.FirstOrDefault(o => string.CompareOrdinal(o.Name, layerName) == 0) != null;
        }

        public bool SafelyRefresh(Overlay overlay)
        {
            bool successfullyRefreshed = false;

            try
            {
                this.Refresh(overlay);
                successfullyRefreshed = true;
            }
            catch (Exception ex)
            {
               
            }

            return successfullyRefreshed;
        }

        public bool SafelyRefresh()
        {
            bool successfullyRefreshed = false;

            try
            {
                this.Refresh();
                successfullyRefreshed = true;
            }
            catch (Exception ex)
            {
               
            }

            return successfullyRefreshed;
        }

        public bool SafelyRefresh(IEnumerable<Overlay> overlays)
        {
            bool successfullyRefreshed = false;

            try
            {
                this.Refresh(overlays);
                successfullyRefreshed = true;
            }
            catch (Exception ex)
            {
               
            }

            return successfullyRefreshed;
        }

        public IEnumerable<FeatureLayer> GetFeatureLayers()
        {
            return GetFeatureLayers(null);
        }

        /// <summary>
        /// This method will return all fature layers on the map, including the EditOverlay.EditShapesLayer
        /// </summary>
        /// <param name="overlaysToExclude"></param>
        /// <returns></returns>
        public IEnumerable<FeatureLayer> GetFeatureLayers(IEnumerable<string> overlaysToExclude)
        {
            var featureLayers = new Collection<FeatureLayer>();
            IEnumerable<LayerOverlay> layerOverlays = GetLayerOverlays(Overlays);

            if (overlaysToExclude != null)
                layerOverlays = layerOverlays.Where(lo => lo != null && !overlaysToExclude.Contains(lo.Name));
            else
                layerOverlays = layerOverlays.Where(lo => lo != null);

            var _featureLayers = GetFeatureLayersFromOverlays(layerOverlays);

            foreach (var featureLayer in _featureLayers)
            {
                featureLayers.Add(featureLayer);
            }
            
            featureLayers.Add(EditOverlay.EditShapesLayer);
            return featureLayers;
        }

        public IEnumerable<Layer> GetLayers(IEnumerable<LayerOverlay> overlays)
        {
            var layers = overlays
                .SelectMany(l => l.Layers)
                .Select(lyr => lyr)
                .Where(fl => fl != null);

            return layers;
        }

        public IEnumerable<LayerOverlay> GetLayerOverlays(IEnumerable<Overlay> overlays)
        {
            var layerOverlays = overlays.Where(o => !o.IsEmpty && o.IsVisible).OfType<LayerOverlay>();
            return layerOverlays;
        }

        public IEnumerable<FeatureLayer> GetFeatureLayersFromOverlays(IEnumerable<LayerOverlay> overlays)
        {
            var featureLayers = overlays.SelectMany(l => l.Layers).OfType<FeatureLayer>().Where(fl => fl.IsVisible);
            return featureLayers;
        }
        #endregion
    }

    public class ViewportExtentChangedEventArgs : EventArgs
    {
        public RectangleShape ViewportExtent { get; set; }
        public double ViewportResolution { get; set; }
    }
}
